import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

public class Sph_parallel_2 {
  public static void main(String[] args) {

    // ガウス関数 exp(-x^2 / h^2)

    // 定数
    final int Ball_Num = 1600; // 粒子数
    final int N = 80; // 繰り返し回数
    final int Xgrid = 400; // 格子数 n^2
    final int Ygrid = 200; 

    int number = 1;

    double dt = 0.1; // 時間幅
    double dx = 0.1; // 格子幅
    double dy = 0.1; // 格子幅

    double t = 0.0;
    double gamma = 1.4; // 比熱比 (圧力計算)
    double h = 2.0; // ガウス関数の影響半径 (密度計算)
    double r = 10.0; // 模型の半径

    // ボール
    Ball[] ball = new Ball[Ball_Num];
    Ball[] print = new Ball[Ball_Num * N];

    // グリッド
    double[][] density = new double[Xgrid][Ygrid]; // グリッド格子点の密度
    double[][] pressure = new double[Xgrid][Ygrid]; // グリッド格子点の圧力
    double[][] dPressureX = new double[Xgrid][Ygrid]; // ∇Px (x成分の差分化)
    double[][] dPressureY = new double[Xgrid][Ygrid]; // ∇Py (y成分の差分化)

    // ルンゲクッタ用係数
    double k1[] = new double[2], k2[] = new double[2], k3[] = new double[2], k4[] = new double[2];

    // ボールの近傍格子点
    int ballGridX;
    int ballGridY;

    // ボールの平均密度
    double ballDensity = 0.0;

    // ボールの平均圧力
    double ballPressureX = 0.0;
    double ballPressureY = 0.0;

    // 初期値
    double m = 1.0; // 粒子の質量
    double v0 = 5.0;
    double theta = 15;

    double v0x = v0 * Math.cos(Math.toRadians(theta)); // 粒子のx軸方向初速度
    double v0y = v0 * Math.sin(Math.toRadians(theta)); // 粒子のy軸方向初速度
    

    // 物体の中の密度
    double den;

    // 格子点
    double xa;
    double ya;

    // 接点
    double x0;
    double y0;

    // 対称点
    double xb;
    double yb;

    // 対称点からの近傍格子点座標
    int GridX;
    int GridY;

    // 配列初期化
    setBall(ball, Ball_Num);
    setBall(print, Ball_Num * N);

    setDensity(density, Xgrid, Ygrid);

    // ボール初期設定
    setBallPlace(ball, dx, dy, v0x, v0y, Ball_Num, r, m);

    // printBall(ball);

    // System.out.println();

    // ---------------------計算開始------------------------

    for (int i = 0; i < N; i++) {

      // ----------------格子点の密度計算部分-----------------
      try {
        FileWriter file1 = new FileWriter(
            "/Users/motoki/oithomes/java/kadai/java21/conflict/case6/parallel/den" + i + ".txt");
        PrintWriter pw1 = new PrintWriter(new BufferedWriter(file1));

        FileWriter file2 = new FileWriter(
            "/Users/motoki/oithomes/java/kadai/java21/conflict/case6/parallel/press" + i + ".txt");
        PrintWriter pw2 = new PrintWriter(new BufferedWriter(file2));

        for (int k = 0; k < Xgrid; k++) {
          for (int l = 0; l < Ygrid; l++) {

            density[k][l] = 0.0;

            // 格子点座標
            xa = k * dx;
            ya = l * dy;

            // 粒子からの格子点密度計算
            // 物体の外
            if (JudgePosition(xa, ya, r) == false) {
              // System.out.printf("%.1f %.1f\n", xa, ya);
              density[k][l] = CalcDensity(ball, h, xa, ya);
            }
          }
        }

        for (int k = 0; k < Xgrid; k++) {
          for (int l = 0; l < Ygrid; l++) {
            // 格子点座標
            xa = k * dx;
            ya = l * dy;

            // System.out.printf("%.1f %.1f\n", xa, ya);

            // 粒子からの格子点密度計算
            // 物体の中
            if (JudgePosition(xa, ya, r) == true) {
              den = CalcDensity(ball, h, xa, ya);

              // System.out.printf("%.1f %.1f\n", xa, ya);

              if (JudgeUpper(xa, ya, r) == true) {
                // System.out.printf("%.1f %.1f\n", xa, ya);

                // 格子点座標からの接点
                x0 = ContactPoint_cir(xa, ya, r);
                y0 = fx1(x0, r);

                //System.out.printf("%f %f\n", x0, y0);

                // 格子点座標に対称な点
                xb = SymmetryX(xa, x0);
                yb = SymmetryY(ya, y0);

                //System.out.printf("%f %f\n", xb, yb);

                // 対称な点に近傍な格子点
                GridX = (int) (xb / dx);
                GridY = (int) (yb / dy);

                // 物体内の格子点密度を対称な点の近傍格子点に分配
                if (GridX != Xgrid - 1 && GridY != Ygrid - 1) {
                  // System.out.printf("%d %d\n", GridX, GridY);
                  density[GridX][GridY] += den / 4.0;
                  density[GridX + 1][GridY] += den / 4.0;
                  density[GridX][GridY + 1] += den / 4.0;
                  density[GridX + 1][GridY + 1] += den / 4.0;
                }
              } else {
                // System.out.printf("%.1f %.1f\n", xa, ya);

                // 格子点座標からの接点
                x0 = ContactPoint_line(xa, ya, r);
                y0 = fx2(x0, r);

                // System.out.printf("%f %f\n", x0, y0);

                // 格子点座標に対称な点
                xb = SymmetryX(xa, x0);
                yb = SymmetryY(ya, y0);

                // System.out.printf("%f %f\n", xb, yb);

                // 対称な点に近傍な格子点
                GridX = (int) (xb / dx);
                GridY = (int) (yb / dy);

                // 物体内の格子点密度を対称な点の近傍格子点に分配
                if (GridX != Xgrid - 1 && GridY != Ygrid - 1) {
                  // System.out.printf("%d %d\n", GridX, GridY);
                  density[GridX][GridY] += den / 4.0;
                  density[GridX + 1][GridY] += den / 4.0;
                  density[GridX][GridY + 1] += den / 4.0;
                  density[GridX + 1][GridY + 1] += den / 4.0;
                }
              }
            }
          }
        }

        for (int k = 0; k < Xgrid; k++) {
          for (int l = 0; l < Ygrid; l++) {
            // 格子点座標
            xa = k * dx;
            ya = l * dy;

            // 圧力計算
            pressure[k][l] = Math.pow(density[k][l], gamma);

            String str1 = String.format("%.1f %.1f %.5f", xa, ya, density[k][l]);
            pw1.println(str1);

            String str2 = String.format("%.1f %.1f %.5f", xa, ya, pressure[k][l]);
            pw2.println(str2);
          }
        }

        // 格子点の圧力を差分化
        for (int k = 1; k < Xgrid - 1; k++) {
          for (int l = 1; l < Ygrid - 1; l++) {
            dPressureX[k][l] = (pressure[k + 1][l] - pressure[k - 1][l]) / (2.0 * dx);
            dPressureY[k][l] = (pressure[k][l + 1] - pressure[k][l - 1]) / (2.0 * dy);
          }

        }
        pw1.close();
        pw2.close();
      } catch (IOException ex) {
        ex.printStackTrace();
      }

      // ----------------粒子の軌跡計算部分-----------------
      for (int j = 0; j < Ball_Num; j++) {

        // 近傍格子点探索
        ballGridX = (int) (ball[j].x / dx);
        ballGridY = (int) (ball[j].y / dy);

        if (ballGridX >= (Xgrid - 1) || ballGridY >= (Ygrid - 1) || ballGridX <= -1 || ballGridY <= -1) {
          ball[j].x += ball[j].vx * dt;
          ball[j].y += ball[j].vy * dt;

          print[i + (j * N)].x = ball[j].x;
          print[i + (j * N)].y = ball[j].y;
        } else {

          print[i + (j * N)].x = ball[j].x;
          print[i + (j * N)].y = ball[j].y;

          // System.out.printf("%d %.5f %.5f\n", j, ball[j].x, ball[j].y);

          // 確認用
          // System.out.println("(" + ballGridX + ", " + ballGridY + ") : " + ball[i].x +
          // ", " + ball[i].y);

          // 平均密度計算
          // System.out.printf("%d %d\n", ballGridX, ballGridY);
          ballDensity = (density[ballGridX][ballGridY] + density[ballGridX + 1][ballGridY]
              + density[ballGridX][ballGridY + 1] + density[ballGridX + 1][ballGridY + 1]) / 4.0;

          // 平均圧力計算: x軸
          ballPressureX = (dPressureX[ballGridX][ballGridY] + dPressureX[ballGridX + 1][ballGridY]
              + dPressureX[ballGridX][ballGridY + 1] + dPressureX[ballGridX + 1][ballGridY + 1]) / 4.0;

          // 平均圧力計算: y軸
          ballPressureY = (dPressureY[ballGridX][ballGridY] + dPressureY[ballGridX + 1][ballGridY]
              + dPressureY[ballGridX][ballGridY + 1] + dPressureY[ballGridX + 1][ballGridY + 1]) / 4.0;

          // 初回の移動
          if (i == 0) {
            ball[j].x += ball[j].vx * dt;
            ball[j].y += ball[j].vy * dt;

            // 以降の移動
          } else {
            // --------------------------------X方向移動------------------------------------------
            k1[0] = dt * f1(t, ball[j].x, ball[j].vx, ballDensity);
            k1[1] = dt * f2(t, ball[j].x, ball[j].vx, ballDensity, ballPressureX);

            k2[0] = dt * f1(t + dt / 2.0, ball[j].x + k1[0] / 2.0, ball[j].vx + k1[1] / 2.0, ballDensity);
            k2[1] = dt
                * f2(t + dt / 2.0, ball[j].x + k1[0] / 2.0, ball[j].vx + k1[1] / 2.0, ballDensity, ballPressureX);

            k3[0] = dt * f1(t + dt / 2.0, ball[j].x + k2[0] / 2.0, ball[j].vx + k2[1] / 2.0, ballDensity);
            k3[1] = dt
                * f2(t + dt / 2.0, ball[j].x + k2[0] / 2.0, ball[j].vx + k2[1] / 2.0, ballDensity, ballPressureX);

            k4[0] = dt * f1(t + dt, ball[j].x + k3[0], ball[j].vx + k3[1], ballDensity);
            k4[1] = dt * f2(t + dt, ball[j].x + k3[0], ball[j].vx + k3[1], ballDensity, ballPressureX);

            ball[j].x = ball[j].x + (k1[0] + 2.0 * k2[0] + 2.0 * k3[0] + k4[0]) / 6.0;
            ball[j].vx = ball[j].vx + (k1[1] + 2.0 * k2[1] + 2.0 * k3[1] + k4[1]) / 6.0;

            // --------------------------------Y方向移動------------------------------------------

            k1[0] = dt * f1(t, ball[j].y, ball[j].vy, ballDensity);
            k1[1] = dt * f2(t, ball[j].y, ball[j].vy, ballDensity, ballPressureY);

            k2[0] = dt * f1(t + dt / 2.0, ball[j].y + k1[0] / 2.0, ball[j].vy + k1[1] / 2.0, ballDensity);
            k2[1] = dt
                * f2(t + dt / 2.0, ball[j].y + k1[0] / 2.0, ball[j].vy + k1[1] / 2.0, ballDensity, ballPressureY);

            k3[0] = dt * f1(t + dt / 2.0, ball[j].y + k2[0] / 2.0, ball[j].vy + k2[1] / 2.0, ballDensity);
            k3[1] = dt
                * f2(t + dt / 2.0, ball[j].y + k2[0] / 2.0, ball[j].vy + k2[1] / 2.0, ballDensity, ballPressureY);

            k4[0] = dt * f1(t + dt, ball[j].y + k3[0], ball[j].vy + k3[1], ballDensity);
            k4[1] = dt * f2(t + dt, ball[j].y + k3[0], ball[j].vy + k3[1], ballDensity, ballPressureY);

            ball[j].y = ball[j].y + (k1[0] + 2.0 * k2[0] + 2.0 * k3[0] + k4[0]) / 6.0;
            ball[j].vy = ball[j].vy + (k1[1] + 2.0 * k2[1] + 2.0 * k3[1] + k4[1]) / 6.0;
          }
        }
      }
      t = t + dt;
      System.out.println(number++);
    }

    try {
      // 板
      FileWriter file_w1 = new FileWriter("/Users/motoki/oithomes/java/kadai/java21/conflict/case6/parallel/dataw.txt");
      PrintWriter pw_w1 = new PrintWriter(new BufferedWriter(file_w1));

      r = 10.0;

      double x = 20.0;
      double y;

      while (x <= 30.0) {
        y = Math.sqrt(Math.pow(r, 2) - Math.pow(x - 25.0, 2)) + 1.0;
        pw_w1.println(x + " " + y);

        x += 0.1;
      }

      x = 30.0;
      y = Math.sqrt(Math.pow(r, 2) - Math.pow(x - 25.0, 2)) + 1.0;

      while (x >= 20.0) {
        pw_w1.println(x + " " + y);
        x -= 0.01;
      }

      pw_w1.close();

      // 粒子
      int num = 0;
      for (int i = 0; i < Ball_Num; i++) {
        FileWriter file = new FileWriter(
            "/Users/motoki/oithomes/java/kadai/java21/conflict/case6/parallel/ball" + num++ + ".txt");
        PrintWriter pw = new PrintWriter(new BufferedWriter(file));
        for (int j = i * N; j < N + i * N; j++) {
          String str = String.format("%.5f %.5f", print[j].x, print[j].y);
          pw.println(str);

          if(j == (N - 1 + i * N)){
            //System.out.println(j + " " + (N - 1 + i * N));
            //System.out.printf("%.5f %.5f\n", print[j].x, print[j].y);
          }
        }
        pw.close();
      }
    } catch (IOException ex) {
      ex.printStackTrace();
    }

  }

  // System.out.println();

  // ------------メソッド-----------------

  // グリッドの密度計算
  public static double CalcDensity(Ball[] ball, double h, double x, double y) {

    double sum = 0; // 密度

    for (int j = 0; j < ball.length; j++) {

      Ball nearBall = new Ball(); // 近傍粒子
      nearBall = ball[j];

      double diff = Math.sqrt(Math.pow(nearBall.x - x, 2) + Math.pow(nearBall.y - y, 2));
      double r2 = Math.pow(diff, 2);

      sum += ball[j].m * Math.exp(-r2 / h);
    }
    return sum;
  }

  // ボール配列初期化
  public static void setBall(Ball[] ball, int n) {
    for (int i = 0; i < n; i++) {
      ball[i] = new Ball();
    }
  }

  // 現在の格子点が物体中かどうかを判定
  public static boolean JudgePosition(double xa, double ya, double r) {
    if (fx2(xa, r) < ya && ya < fx1(xa, r)) {
      return true;
    } else {
      return false;
    }
  }

  // ボール初期設定
  public static void setBallPlace(Ball[] ball, double dx, double dy, double v0x, double v0y, int Ball_Num, double r, double m) {

    int N = (int) Math.sqrt(Ball_Num);

    double xa;
    double ya;

    double xmin = 2.0;
    double xmax = 30.0;
    double ymin = 2.0;
    double ymax = 18.0;

    for (int i = 0; i < N; i++) {
      for (int j = 0; j < N; j++) {
        xa = xmin + i * (xmax - xmin) / (N - 1);
        ya = ymin + j * (ymax - ymin) / (N - 1);

        if(JudgePosition(xa, ya, r)){
          ball[(i * N) + j].x = xmin + i * (xmax - xmin) / (N - 1);
          ball[(i * N) + j].y = ymin + j * (ymax - ymin) / (N - 1);

          ball[(i * N) + j].vx = 0.0;
          ball[(i * N) + j].vy = 0.0;

          ball[(i * N) + j].m = 0.0;
        } else {
          ball[(i * N) + j].x = xmin + i * (xmax - xmin) / (N - 1);
          ball[(i * N) + j].y = ymin + j * (ymax - ymin) / (N - 1);

          ball[(i * N) + j].vx = v0x;
          ball[(i * N) + j].vy = v0y;

          ball[(i * N) + j].m = m;
        }

        
      }
    }

  }

  // 密度配列初期化
  public static void setDensity(double[][] density, int Xgrid, int Ygrid) {
    for (int i = 0; i < Xgrid; i++) {
      for (int j = 0; j < Ygrid; j++) {
        density[i][j] = 0.0;
      }
    }
  }

  // ボール配列確認
  public static void printBall(Ball[] ball) {
    for (int i = 0; i < ball.length; i++) {
      System.out.println(ball[i].x + " " + ball[i].y);
    }
  }

  // 密度配列確認
  public static void printDensity(double[][] density, int Xgrid, int Ygrid, double dx, double dy) {
    for (int i = 0; i < Xgrid; i++) {
      for (int j = 0; j < Ygrid; j++) {
        System.out.printf("%.1f %.1f %.5f\n", i * dx, j * dy, density[i][j]);
      }
    }
  }

  // ルンゲクッタ法: 速度
  public static double f1(double t, double x, double v, double density) {
    if (density < 0.01) {
      return 0.0;
    } else {
      return v;
    }
  }

  // ルンゲクッタ法: 加速度
  public static double f2(double t, double x, double v, double density, double DelPressure) {
    if (density < 0.01) {
      return 0.0;
    } else {
      return -(1.0 / density) * DelPressure;
    }
  }

  // 模型 円形
  public static double calcCircle(double x, double r) {
    double y;

    y = Math.sqrt(Math.pow(r, 2) - Math.pow(x - 25.0, 2)) + 1.0;
    // System.out.println(y);
    return y;
  }

  // 模型 直線
  public static double calcLine(double x, double r) {
    double y;
    x = 20.0;
    y = calcCircle(x, r);
    return y;
  }

  // 模型の中点座標計算
  public static double calcMiddlePoint(double y1, double y2) {
    double y;

    y = (y1 + y2) / 2.0;
    return y;
  }

  // 板 円形
  public static double fx1(double x, double r) {
    double y;

    y = Math.sqrt(Math.pow(r, 2) - Math.pow(x - 25.0, 2)) + 1.0;
    return y;
  }

  public static double dfx1(double x, double r) {
    double y;

    y = -(x - 25.0) / Math.sqrt(Math.pow(r, 2) - Math.pow(x - 25.0, 2));
    return y;
  }

  // 板 直線
  public static double fx2(double x, double r) {
    double y = calcCircle(20.0, r);
    return y;
  }

  public static double dfx2(double x, double r) {
    return 0.0;// -------------------------------------
  }

  // 接点を算出 円形
  public static double ContactPoint_cir(double xa, double ya, double r) {
    int f = 0;
    double x0 = 0.0;
    double h = 0.01;
    double a = 0.0;

    //System.out.println(xa + " " + ya);

    if (xa == 25.0) {
      x0 = 25.0;

    } else if (xa < 25.0) {
      while (f == 0) {
        a = ya - fx1(x0, r) + (xa - x0) / dfx1(x0, r);

        //System.out.println(xa + " " + ya + " " + x0 + " " + a + " " + dfx1(x0, r));

        if (a <= 0.0) {
          //System.out.println(xa + " " + ya + " " + x0 + " " + a + " " + dfx1(x0, r));
          f = 1;
        } else {
          x0 += h;
        }
      }
    } else if (xa > 25.0) {
      x0 = 25.1;
      while (f == 0) {
        a = ya - fx1(x0, r) + (xa - x0) / dfx1(x0, r);
        
        if (a >= 0.0) {
          f = 1;
        } else {
          x0 += h;
        }
      }
    }

    return x0;
  }

  // 接点を算出 直線
  public static double ContactPoint_line(double x, double y, double r) {
    int f = 0;
    double x0 = 0.0;
    double h = 0.01;
    double a;

    while (f == 0) {
      a = y - fx2(x0, r) + (x - x0) / dfx2(x0, r);
      if (a <= 0.0) {
        f = 1;
      } else {
        x0 += h;
      }
    }

    return x0;
  }

  // 対称な点を算出: x
  public static double SymmetryX(double xa, double x0) {
    return xa + 2 * (x0 - xa);
  }

  // 対称な点を算出: y
  public static double SymmetryY(double ya, double y0) {
    return ya + 2 * (y0 - ya);
  }

  // 格子点が中点より上か
  public static boolean JudgeUpper(double xa, double ya, double r) {

    // System.out.println(xa + " " + calcMiddlePoint(calcLine(xa, r), calcCircle(xa,
    // r)));

    if (ya >= calcMiddlePoint(calcLine(xa, r), calcCircle(xa, r))) {
      return true;
    } else {
      return false;
    }
  }

}
